package org.orekit.celestrak;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;

import org.orekit.propagation.analytical.tle.TLE;

public class CelestrakTLE extends TLE {

	private static final long serialVersionUID = 1L;
	
	/**
	 * Object name
	 */
	private String name;
	
	/**
	 * Line 1
	 */
	private String line1;
	
	/**
	 * Line 2
	 */
	private String line2;
	
	/**
	 * Get object name
	 */
	public String getName() {
		return this.name;
	}
	
	/**
	 * Get TLE object
	 * @return TLE
	 */
	public TLE getTLE() {
		return new TLE(this.line1, this.line2);
	}
	
	/**
	 * Basic constructor
	 * @param name object name
	 * @param line1 TLE lin1
	 * @param line2 TLE line2
	 */
	public CelestrakTLE(String name, String line1, String line2) {
		super(line1, line2);
		this.name = name;
		this.line1 = line1;
		this.line2 = line2;
	}

	/**
     * Update TLE database
	 * @throws IOException 
	 * @throws InterruptedException 
     */
	public static void update(String[] TLERetriever) 
			throws IOException, InterruptedException{
		
        System.out.println(">> Updating TLE database from Celestrak ...");
        Process p = Runtime.getRuntime().exec(TLERetriever);        		
        p.waitFor();
        System.out.println("TLE database successfully updated. \n");        	
    	
    }
	
	/**
     * Update and copy database into project resources
	 * @throws IOException 
	 * @throws InterruptedException 
     */
	public static void updateResources(String[] TLERetriever, Path resourcesPath) 
			throws IOException, InterruptedException{

		// Update TLE database from Celestrak 
		update(TLERetriever);
		
		System.out.println(">> Copying TLE data into " + resourcesPath);
		File resourcesDirectory = resourcesPath.toFile();
		String[] files = resourcesDirectory.list();
		
		/* If resources folder is not empty, transfer TLE database 
		using transfer directory temp */
		if (files.length > 0) {
			
			// Create temp directory and resources folder location
			Path tempDirectory = Paths.get(resourcesPath.getParent().toString() + "\\temp");
			if (tempDirectory.toFile().exists()) tempDirectory.toFile().delete();
			tempDirectory.toFile().mkdir();
			
			// Move each resources file into temp to empty resources folder
			for (String f : files) {
				
				Files.move(
						Paths.get(resourcesPath + "\\" + f), 
						Paths.get(tempDirectory + "\\" + f), 
						StandardCopyOption.REPLACE_EXISTING);
			
			}
			
			// Copy TLE database file into temp
			Files.copy(
					Paths.get("d:\\Users\\mlargeau\\Documents\\My TLEs\\CelesTrak\\catalog.tle"), 
					Paths.get(tempDirectory + "\\catalog.tle"), 
					StandardCopyOption.REPLACE_EXISTING);

			// Move temp content into resources
			for (String f : tempDirectory.toFile().list()) {
				
				Files.move(
						Paths.get(tempDirectory + "\\" + f), 
						Paths.get(resourcesPath + "\\" + f), 
						StandardCopyOption.REPLACE_EXISTING);
				
			}	
			
			// Delete temp directory
			Files.delete(tempDirectory);
			
		}
		
		// If resources folder is empty, copy TLE dabatabase
		else {
			
			Files.copy(
					Paths.get("d:\\Users\\mlargeau\\Documents\\My TLEs\\CelesTrak\\catalog.tle"), 
					Paths.get(resourcesPath + "\\catalog.tle"), 
					StandardCopyOption.REPLACE_EXISTING);
			
		}
		
	System.out.println("TLE data transferred successfully.");
    	
    }
	
	
	/**
     * Load debris input from Celestrak catalog.tle file
	 * @throws IOException 
     */
	public static List<CelestrakTLE> parseTLE(String TLEDataPath, Integer targetLimit) 
			throws IOException{
		
	    System.out.println(">> Reading TLE input file " + TLEDataPath);
	    
	    // Counting number of lines and checking
	    List<String> allLines = Files.readAllLines(Paths.get(TLEDataPath));
	    if (allLines.size()%3 != 0) throw new IllegalStateException(
	    		"The number of lines in TLE catalog file must be a multiple of 3."); 
	    
	    // Handle limit on number of objects as prompted by user
		if (targetLimit != null){
			
	        if (targetLimit > allLines.size()/3) throw new IllegalStateException(
	        		"Number of targets required (" + targetLimit + ") "
	        				+ "cannot be higher than number of objects available "
	        				+ "(" + allLines.size()/3); 
	        
		} else {
			
			targetLimit = allLines.size()/3;
			
		}
			
		List<CelestrakTLE> objectsList = new ArrayList<CelestrakTLE>(); 
		for (int i = 0; i <= 3*targetLimit -1; i+=3) {

			objectsList.add(new CelestrakTLE(
					allLines.get(i), 
					allLines.get(i+1), 
					allLines.get(i+2)));
			
		}
		
	    System.out.println("TLE catalog file read successfully. \n");
		return objectsList;
	    
	}

}
