package org.orekit.detection;

import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.util.FastMath;
import org.orekit.geometry.fov.FieldOfView;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.events.AbstractDetector;
import org.orekit.propagation.events.VisibilityTrigger;
import org.orekit.propagation.events.handlers.EventHandler;
import org.orekit.propagation.events.handlers.StopOnIncreasing;
import org.orekit.utils.PVCoordinatesProvider;

public class RadarFieldOfViewDetector extends AbstractDetector<RadarFieldOfViewDetector> {
	
	/**
	 * Radar name
	 */
	private String radarName;
	
	/**
	 * Radar range
	 */
	private double radarRange;
	
	/** 
	 * Radar field of view
	 */
	private final FieldOfView fov;
	
	/**
	 * Target name
	 */
	private final String targetName;
	
	/**
	 * Target position/velocity provider 
	 */
	private final PVCoordinatesProvider targetPVProvider;
	
	/**
	 * Target radius
	 */
	private final double radiusTarget;
	
	/**
	 * Visibility trigger
	 */
	private final VisibilityTrigger trigger;
	
	/**
	 * Max check duration for radar visibility event 
	 */
	private static double DEFAULT_MAXCHECK_RADAR = 1.0;
	
	public FieldOfView getFOV() {
		return this.fov;
	}
	
	/**
	 * Get Radar name
	 */
	public String getRadarName() {
		return this.radarName;
	}
	
	/**
	 * Get target Name
	 */
	public String getTargetName() {
		return this.targetName;
	}
	
	/**
	 * Get targetPVProvider
	 * @return PV coordinates provider
	 */
	public PVCoordinatesProvider getTargetPVProvider() {
		return this.targetPVProvider;
	}
	
	/** Public constructor with basic parameters.
	 * </p>
	 * @param pvTarget target position/velocity provider
	 * @param radarRange radar range 
	 * @param fov radar field of view
	 */	
	public RadarFieldOfViewDetector(String name, PVCoordinatesProvider pvTarget, String targetName, 
			double radarRange, FieldOfView fov) {
		this(DEFAULT_MAXCHECK_RADAR, DEFAULT_THRESHOLD, DEFAULT_MAX_ITER,
	             new StopOnIncreasing<RadarFieldOfViewDetector>(),
	             name, pvTarget, targetName, 0.0, radarRange, fov);
	}
	
	/** Private constructor with full parameters.
	 * </p>
	 * @param maxCheck
	 * @param threshold
	 * @param maxIter
	 * @param handler
	 * @param pvTarget
	 * @param radiusTarget
	 * @param radarRange
	 * @param fov
	 */
	private RadarFieldOfViewDetector(final double maxCheck, final double threshold, final int maxIter,
                           final EventHandler<? super RadarFieldOfViewDetector> handler, 
                           String name, PVCoordinatesProvider pvTarget, String targetName, double radiusTarget, 
                           double radarRange, FieldOfView fov) {
	super(maxCheck, threshold, maxIter, handler);
	this.radarName = name;
	this.targetPVProvider = pvTarget;
	this.radarRange = radarRange;
	this.fov = fov;
	this.targetName = targetName;
	this.radiusTarget = radiusTarget;
	this.trigger = VisibilityTrigger.VISIBLE_AS_SOON_AS_PARTIALLY_IN_FOV;
		
	}
	
	/** {@inheritDoc} */
    @Override
    protected RadarFieldOfViewDetector create(final double newMaxCheck, final double newThreshold,
                                         final int newMaxIter,
                                         final EventHandler<? super RadarFieldOfViewDetector> newHandler) {
        return new RadarFieldOfViewDetector(newMaxCheck, newThreshold, newMaxIter, newHandler,
                                       radarName, targetPVProvider, targetName, radiusTarget, radarRange, fov);
    }

    /** State monitoring function for event detection
     * </p>
     * @param s spacecraft current state
     */
	public double g(final SpacecraftState s) {

        // Get line of sight in spacecraft frame
        final Vector3D targetPosInert =
                targetPVProvider.getPVCoordinates(s.getDate(), s.getFrame()).getPosition();
        final Vector3D lineOfSightSC = s.toTransform().transformPosition(targetPosInert);

        final double range = this.radarRange - lineOfSightSC.getNorm();
        final double angularRadius = FastMath.asin(radiusTarget / lineOfSightSC.getNorm());
        final double offset = fov.offsetFromBoundary(lineOfSightSC, angularRadius, trigger);
        
        // If target is in infinite field of view, take range into account 
        if (offset < 0) return offset*range;	        	
        
        // Otherwise, watch offset only
        return offset;	  
        
    }
	
}
