package org.orekit.detection;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.hipparchus.ode.events.Action;
import org.hipparchus.util.FastMath;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.errors.OrekitException;
import org.orekit.geometry.fov.CircularFieldOfView;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.propagation.events.EventsLogger;
import org.orekit.spacecraft.SpaceBorneRadar;

public class VisibilityTLEPropagator {	
	
	/**
	 * Spaceborne radar carrier propagator
	 */
	private TLEPropagator carrierPropagator;
	
	/**
	 * Events logger
	 */
	private EventsLogger logger; 
	
	/**
	 * List of untreated targets. 
	 * TLE propagator building may fail if no convergence is achieved.
	 */
	private List<CelestrakTLE> untreatedTargets;
	
	/**
	 * Get carrier TLE propagator
	 * @return carrier propagator
	 */
	public TLEPropagator getTLEPropagator() {
		return this.carrierPropagator;
	}
	
	/**
	 * Get the list of untreated targets
	 * @return list of Celestrak TLE objects
	 */
	public List<CelestrakTLE> getUntreatedTargets() {
		return this.untreatedTargets;
	}
	
	/**
	 * Get events logger
	 * @return EventsLogger
	 */
	public EventsLogger getLogger() {
		return logger;
	}
	
	/** Compute radar visibilities
     * @param args program arguments
     * @throws IOException
	 * @throws URISyntaxException 
	 * @throws InterruptedException 
     */
    public VisibilityTLEPropagator(
    		SpaceBorneRadar carrier, 
    		List<CelestrakTLE> objectsList) 
			throws IOException, URISyntaxException, InterruptedException {                        
    	
			Logger warnLogger = Logger.getLogger(VisibilityTLEPropagator.class.getName());
			warnLogger.setLevel(Level.WARNING);  
			
			//Initializing propagators and event detectors  
			final TLEPropagator carrierPropagator = TLEPropagator
					.selectExtrapolator(carrier.getTLE());
                           
            carrierPropagator.setAttitudeProvider(carrier.getAttitudeProvider());
            
            // Build field of view detector for each debris
            final List<CelestrakTLE> untreatedDebrisList = new ArrayList<>();
            final EventsLogger logger = new EventsLogger();
            for (CelestrakTLE object : objectsList) {

            	TLEPropagator currentPropagator = null;
            	try {           	                                 
                	currentPropagator = TLEPropagator
        					.selectExtrapolator(object.getTLE()); 	                		
            	}
            	catch (OrekitException e){
            	
            		untreatedDebrisList.add(object);
            		warnLogger.warning("The object " + object.getName() + " will not be treated." 
            				+ " The propagator could not be built : ");
            		e.printStackTrace();
            		continue;
            		
            	}
	                
                // Basic radar field of view detector
                RadarFieldOfViewDetector currentRadarDetector = 
                		new RadarFieldOfViewDetector(
                				carrier.getRadar().getName(),
                				currentPropagator.getPvProvider(), 
                				object.getName(),
                				carrier.getRadar().getRange(),
                				new CircularFieldOfView(
                						carrier.getRadar().getOrientation(), 
                						FastMath.toRadians(carrier.getRadar().getHalfAperture()), 
                						0.0))
                		.withHandler((s, detector, increasing) -> {
                			long threadId = Thread.currentThread().threadId();
                			if (increasing) {                    				
                				System.out.println(threadId + " : " + s.getDate() + " : EXIT : " + detector.getTargetName());
                			} else 
                				System.out.println(threadId + " : " + s.getDate() + " : ENTER : " + detector.getTargetName());
                			return Action.CONTINUE;});

                // Add field of view detector to SBR TLE propagator with step handler
                carrierPropagator.addEventDetector(currentRadarDetector);               
                
                // Add event logger for further analyses
                carrierPropagator.addEventDetector(logger.monitorDetector(currentRadarDetector));
                
            }                     
            
            this.carrierPropagator = carrierPropagator;
        	this.untreatedTargets = untreatedDebrisList;
        	this.logger = logger;
            
    }
        
}

