package org.orekit.logger;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.util.FastMath;
import org.orekit.detection.RadarFieldOfViewDetector;
import org.orekit.frames.FramesFactory;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.events.EventsLogger.LoggedEvent;
import org.orekit.propagation.events.VisibilityTrigger;
import org.orekit.time.AbsoluteDate;
import org.orekit.utils.PVCoordinatesProvider;

public class RadarVisibilityEventLogger {
	
	/**
	 * Generate events chronogram from logger 
	 * @param logger event logger
	 * @param startDate propagation start date
	 * @param endDate propagation end date
	 * @throws IOException
	 */
    public static void logEvents(List<LoggedEvent> loggedEvents, 
    		AbsoluteDate startDate, 
    		AbsoluteDate endDate) throws IOException {
    	
    	// Events start/end dates
    	Map<String, AbsoluteDate> startPerTarget = new HashMap<>();
    	Map<String, AbsoluteDate> endPerTarget = new HashMap<>();
    	
        // Output file
        String chronoOutputFileName = "chonogram.csv";
        BufferedWriter chronoWriter = new BufferedWriter(new FileWriter(chronoOutputFileName));
        String chronoTitle = "RADAR VISIBILITY EVENTS CHRONOGRAM";
        String chronoHeader = "Event; Date (UTC); Radar ID; Target; "
        		+ "Range (m); Offset (rad); Duration (s); Gap (s)";
        
        chronoWriter.write(chronoTitle + "\n\n" 
        		+ "Propagation parameters : \n" 
        		+ "  Start date : " + startDate + "\n"
        		+ "  End date : " + endDate + "\n\n" 
        		+ chronoHeader + "\n\n");
    	
        // Loop on events
    	for (LoggedEvent event : loggedEvents) {
            
    		// State
    		SpacecraftState s = event.getState();
    		
    		// Associated detector
    		RadarFieldOfViewDetector currentDetector = (RadarFieldOfViewDetector)
    				event.getEventDetector();
    		PVCoordinatesProvider currentPVProvider = currentDetector.getTargetPVProvider();
    		String targetName = currentDetector.getTargetName();
    		
    		// Compute range               	
    		Vector3D deltaPos = s.getPVCoordinates().getPosition()
    				.subtract(currentPVProvider.getPVCoordinates(s.getDate(), FramesFactory.getTEME())
    						.getPosition());
    		double range = deltaPos.getNorm();
    		
    		// Compute angle between line of sight and field of view closest boundary
    		final Vector3D targetInertPos =
    				currentPVProvider.getPVCoordinates(s.getDate(), s.getFrame()).getPosition();
            final Vector3D targetLocalPos = s.toTransform().transformPosition(targetInertPos);

            final double offset = currentDetector.getFOV()
            		.offsetFromBoundary(targetLocalPos, 0.0, 
            				VisibilityTrigger.VISIBLE_AS_SOON_AS_PARTIALLY_IN_FOV);

    		// Tell whether object enters or exit the field of view and compute duration/gap
    		String eventType = "ENTER";
    		Double duration = null;
    		Double gap = null;
    		
    		if (event.isIncreasing()) { 
    			
    			eventType = "EXIT";
    			endPerTarget.put(targetName, event.getDate());
    			AbsoluteDate eventStartDate = startPerTarget.get(targetName);
    			
    			if (eventStartDate != null) duration = s.getDate()
    					.durationFrom(eventStartDate);
    			else duration = s.getDate().durationFrom(startDate);
    			
    		} else {
    			
    			AbsoluteDate eventEndDate = endPerTarget.get(targetName);
    			
    			if (eventEndDate != null) {
    				gap = event.getDate().durationFrom(eventEndDate);				
    			}
    			else gap = s.getDate().durationFrom(startDate);
    			
    			startPerTarget.put(targetName, event.getDate());
    			
    		}
    			    		
    		// Writing in file	           			
    		try {
    			chronoWriter.write(eventType + "_FIELD_OF_VIEW; "
    					+ s.getDate() + "; "                					
    					+ currentDetector.getRadarName() + "; "
    					+ currentDetector.getTargetName() + "; "
    					+ range + "; "
    					+ FastMath.toDegrees(offset) + "; " 
    					+ duration + "; " 
    					+ gap + "\n");
    		} catch (IOException e) {
    			e.printStackTrace();
    		}     		
    		
    	}
    	
		chronoWriter.close();
		System.out.println("File " + chronoOutputFileName + " generated successfully. \n");	
    	
    }

}
