package org.orekit.synthesis;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.orekit.bodies.CelestialBodyFactory;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.detection.RadarFieldOfViewDetector;
import org.orekit.frames.FramesFactory;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.orbits.Orbit;
import org.orekit.propagation.events.EventsLogger.LoggedEvent;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.ChronologicalComparator;
import org.orekit.utils.Constants;
import org.orekit.utils.TimeStampedPVCoordinates;

public class EventSynthesis {
	
	private static double LEO = Constants.WGS84_EARTH_EQUATORIAL_RADIUS + 2000000.0;
	private static double MEO = Constants.WGS84_EARTH_EQUATORIAL_RADIUS + 20000000.0;
	private static double GEO = Constants.WGS84_EARTH_EQUATORIAL_RADIUS + 36000000.0;
	
    /**
     * Generate statistics on visibility events 
     * (number of occurrences for each target, 
     * max duration in field of view,
     * max gaps between two consecutive occurrences, ...).
     * Results are saved in a txt file.
     * </p>
     * @param logger events logger
     * @param propagationStartDate propagation start date
     * @param propagationEndDate propagation end date
     * @param untreated objects list 
     * @throws IOException
     */
    public static void generate(List<LoggedEvent> events, 
    		AbsoluteDate propagationStartDate,
    		AbsoluteDate propagationEndDate, 
    		int totalObjectsNumber,
    		List<CelestrakTLE> untreatedObjectsList) throws IOException {
    	
        // Synthesis file
        String synthesisOutputFileName = "synthesis.txt";
        String synthesisTitle = "RADAR VISIBILITY EVENTS SYNTHESIS";
        String synthesisHeader = "Target; Full visibility occurrences; "
        		+ "Max. duration inside field of view; "
        		+ "Max. gap between visibilities";
    	
    	// Buffered file writer
    	BufferedWriter synthesisWriter = new BufferedWriter(new FileWriter(synthesisOutputFileName));
    	
    	// Get all radar visibility events
	    Collections.sort(events, new ChronologicalComparator());
	
	    // Compute various statistics
	    Map<String, Integer> countPerTarget = new TreeMap<>();	    
	    Map<String, Double> maxDurationPerTarget = new TreeMap<>();
	    Map<String, Double> maxGapPerTarget = new TreeMap<>();
	    Map<String, AbsoluteDate> startPerTarget = new HashMap<>();
	    Map<String, AbsoluteDate> endPerTarget = new HashMap<>();
	    List<Double> durations = new ArrayList<>();
	    List<Double> gaps = new ArrayList<>();    
	    
	    double countLEO = 0;
	    double countMEO = 0;
	    double countGEO = 0;
	    
	    // Loop on events
	    for (LoggedEvent event : events) {
	    	
	        // Retrieve target name from current event
	        RadarFieldOfViewDetector rd  = (RadarFieldOfViewDetector) event.getEventDetector();
	        String targetName = rd.getTargetName();
	
	        // Count visibility occurrences
	        if (!event.isIncreasing()) {
	        	
	            Integer previousCount = countPerTarget.get(targetName);
	            countPerTarget.put(targetName, previousCount == null ? 1 : previousCount.intValue() + 1);
	            
	            // Target in LEO, MEO or GEO orbit
	            TimeStampedPVCoordinates targetPV = rd.getTargetPVProvider()
	            		.getPVCoordinates(propagationEndDate, FramesFactory.getTEME());
	            Orbit orbit = new KeplerianOrbit(targetPV, FramesFactory.getTEME(), 
	            		CelestialBodyFactory.getEarth().getGM());
	            if (previousCount == null) {
	            	if (orbit.getA() <= LEO) countLEO ++;
	            	else if (orbit.getA() <= MEO) countMEO ++;
	            	else countGEO ++;	            	
	            }
	            
	        }
	
	        // Compute visibility durations and gaps
	        if (!event.isIncreasing()) {
	        	
	            startPerTarget.put(targetName, event.getDate());
	            AbsoluteDate eventEndDate = endPerTarget.get(targetName);
	            
	            // If an exit was previously found
	            if (eventEndDate != null) {
	            	
	                double gap = event.getDate().durationFrom(eventEndDate);
	                gaps.add(gap);
	                                     
	                Double previousGap = maxGapPerTarget.get(targetName);
	                maxGapPerTarget.put(targetName, 
	                		previousGap == null || previousGap < gap ? gap : previousGap);
	                
	            } else maxGapPerTarget.put(targetName, 
	            		event.getDate().durationFrom(propagationStartDate));
	            
	        } else {
	        	
	        	endPerTarget.put(targetName, event.getDate());	            	            
	            AbsoluteDate eventStartDate = startPerTarget.get(targetName);
	            
	            // If only the exit event is present (e.g. at the beginning of the chronogram)
	            if (startPerTarget.get(targetName) == null) {
	            	eventStartDate = propagationStartDate;
	            }
	            
	            // Compute event duration
	            double duration = event.getDate().durationFrom(eventStartDate);
	            
	            // As of the latest version, only take full events durations into account 
	            if (startPerTarget.containsKey(targetName)) {
	                durations.add(duration);
	            }
	            
	            startPerTarget.remove(targetName);
	            Double previousMax = maxDurationPerTarget.get(targetName);
	            maxDurationPerTarget.put(targetName, 
	            		previousMax == null || previousMax < duration ? duration : previousMax);
	            
	        }
	        
	    }
	    
	    // Write results in synthesis file
	    synthesisWriter.write(synthesisTitle + "\n\n" 
	    		+ "Propagation parameters : \n" 
	    		+ "  Start date : " + propagationStartDate + "\n"
	    		+ "  End date : " + propagationEndDate + "\n\n" 
	    		+ synthesisHeader + "\n\n");
	    
	    for (Entry<String, Integer> entry : countPerTarget.entrySet()) { 	    	
	    	String currentKey = entry.getKey();
	    	synthesisWriter.write(currentKey + "; " 
	    			+ entry.getValue() +"; "
	    			+ maxDurationPerTarget.get(currentKey) + "; "
	    			+ maxGapPerTarget.get(currentKey) + "\n");
	    }
	
	    // Targets coverage
	    synthesisWriter.write("\n" + "Coverage : \n");
	    synthesisWriter.write("    Total : " + countPerTarget.size() + "/" 
	    		+ totalObjectsNumber + "\n");
	    synthesisWriter.write("    Orbits : \n"
	    		+ "        LEO : " + countLEO + "\n"
	    		+ "        MEO : " + countMEO + "\n"
	    		+ "        GEO : " + countGEO + "\n");
	    
	    // Write names of untreated debris
	    synthesisWriter.write("\n" + "Untreated debris : \n");
	    for (CelestrakTLE object : untreatedObjectsList) {
	    	synthesisWriter.write(object.getName() + "\n");
	    }
	    
	    // Visibility occurrences spreading
	    List<Double> occurrences = new ArrayList<>();
	    List<Double> nbObjects = new ArrayList<>();
	    Integer maxVisiOccurences = Collections.max(countPerTarget.values());
	    for (double i = 0; i <= maxVisiOccurences; i++) {	
		    Integer objectsCounter = 0;
	    	for (Entry<String, Integer> entry : countPerTarget.entrySet()) { 	    	
		    	if (entry.getValue() == i) objectsCounter ++;
		    }
	    	occurrences.add(i);
	    	nbObjects.add(objectsCounter.doubleValue());
	    }	    
	    
	    File output = new File("objects.png");
	    barChart(occurrences, nbObjects, "Visibility occurrences spreading", 
	    		"'Occurrences'", "'Number of objects'", output);
	    
	    synthesisWriter.close();
	    System.out.println("File " + synthesisOutputFileName + " generated successfully." + "\n");
	    
    }
    
    public static void barChart(List<Double> x, List<Double> y, 
    		String title, String xlabel, String ylabel, File outputFile) throws IOException {
    	
    	final ProcessBuilder pb = new ProcessBuilder("gnuplot").
                redirectOutput(ProcessBuilder.Redirect.INHERIT).
                redirectError(ProcessBuilder.Redirect.INHERIT);
	    pb.environment().remove("XDG_SESSION_TYPE");
	    final Process gnuplot = pb.start();
	    
	    try (PrintStream out = new PrintStream(gnuplot.getOutputStream(), false, StandardCharsets.UTF_8.name())) {
            if (outputFile == null) {
            	out.format(Locale.US, "set terminal qt size %d, %d title 'complex plotter'%n",
                           1000, 1000);
            } else {
            	out.format(Locale.US, "set terminal pngcairo size %d, %d%n", 1000, 1000);        
            	out.format(Locale.US, "  set output '%s'%n", outputFile.getAbsolutePath());
            }

            out.format(Locale.US, "set xrange [%f : %f]%n", Collections.min(x), Collections.max(x));
            out.format(Locale.US, "set yrange [%f : %f]%n", Collections.min(y), Collections.max(y));
            out.format(Locale.US, "  set xlabel %s%n", xlabel);
            out.format(Locale.US, "set xlabel font '%s'%n", "Calibri, 15");
            out.format(Locale.US, "set ylabel %s%n", ylabel);
            out.format(Locale.US, "set ylabel font '%s'%n", "Calibri, 15");
            out.format(Locale.US, "set title '%s'%n", title);
            out.format(Locale.US, "set title font '%s'%n", "Calibri, 20");
            out.format(Locale.US, "set boxwidth 0.5%n");
            out.format(Locale.US, "set style fill solid%n");
            out.format(Locale.US, "set grid%n");
            out.format(Locale.US, "$data <<EOD%n"); // Sending data to gnuplot
            for (int i = 0; i < x.size(); i++) { 	
            	out.format(Locale.US, "%f %f%n", x.get(i), y.get(i));
            }
            out.format(Locale.US, "EOD%n");
            out.format(Locale.US, "plot $data using 1:2 notitle with boxes %n");

            if (outputFile == null) {
            	out.format(Locale.US, "pause mouse close%n");
            	} else {
            		System.out.format(Locale.US, "Output figure written to %s.%n", outputFile.getAbsolutePath());
        	}
    	
	    }
	    
    }

}
