package MultipleTargetsMonitoring;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Locale;

import org.hipparchus.util.FastMath;
import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.LofOffset;
import org.orekit.bodies.CelestialBodyFactory;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.detection.VisibilityTLEPropagator;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.LOFType;
import org.orekit.logger.RadarVisibilityEventLogger;
import org.orekit.spacecraft.Radar;
import org.orekit.spacecraft.SpaceBorneRadar;
import org.orekit.synthesis.EventSynthesis;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScalesFactory;

import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

import SimpleChaserTarget.SimpleChaserTarget;

public class FixedRadarConstantAttitude {
	
	// Limit on number of objects to be loaded 
	private static int OBJECTS_NUMBER = 300;
	
	// Radar input file name
	private static String RADAR_INPUT_FILE = "fixed-radar-constant-attitude.yaml";
	
	// Propagation start date
	private static int YEAR = 2023;
	private static int MONTH = 01;
	private static int DAY = 26;
	private static int HOUR = 03;
	private static int MIN = 35;
	private static double SEC = 45.396;
	
	// Propagation horizon
	private static double PROPAGATION_HORIZON = 3600;
	
	public static void main(final String[] args) 
			throws URISyntaxException, 
					StreamReadException, 
					DatabindException, 
					IOException, 
					InterruptedException {
		
		// Load Orekit data
        final File home = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, "
            		+ "unzip it in %s and rename it 'orekit-data' for this "
            		+ "tutorial to work%n",
                    "orekit-data-master.zip", 
                    "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                    home.getAbsolutePath());
            System.exit(1);
        }
        
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));  
		       
        // Time display
        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("uuuu/MM/dd HH:mm:ss.SSS");
		        
        // Parse radar inputs from file
        System.out.println(dtf.format(LocalDateTime.now()) + " Reading radar input file");
        
        final URL url = SimpleChaserTarget.class.getResource("/" + RADAR_INPUT_FILE);
        if (url == null) {
            System.err.println("Resource " + RADAR_INPUT_FILE + " not found.");
            System.exit(1);
        }

		final File inputFile  = new File(url.toURI().getPath());
        
        // Read radar input parameters from parsed file
        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        final Radar radar = mapper.readValue(inputFile, Radar.class);
        radar.setOrientation(radar.getU1(), radar.getU2(), radar.getU3());
        
        System.out.println(dtf.format(LocalDateTime.now()) 
        		+ " File " + inputFile.getName() 
        		+ " parsed successfully. Properties : \n" 
        		+ "  Name : " + radar.getName() + "\n"
                + "  Range : " + radar.getRange() + " m \n"
                + "  Half aperture : " + radar.getHalfAperture() + " deg \n"
                + "  Orientation : " + radar.getOrientation() + " \n");   
		
        // Parse radar carrier parameters    
        final URL carrierURL = SimpleChaserTarget.class.getResource("/carrier.tle");
        CelestrakTLE carrierTLE = CelestrakTLE.parseTLE(
        		carrierURL.getPath().replace("/", "\\").substring(1), null).get(0);
        
        Double mu = CelestialBodyFactory.getEarth().getGM();
        Double carrierN = carrierTLE.getMeanMotion();
        Double carrierA = FastMath.cbrt(mu/(carrierN*carrierN));
        System.out.println("Carrier semi-major axis from TLE input : " 
                + carrierA + " m. \n");
        
        // Align satellite frame with TNW frame
        final AttitudeProvider attitudeLaw = new LofOffset(FramesFactory.getTEME(), LOFType.TNW);
        
        // Radar carrier
        SpaceBorneRadar carrier = 
        		new SpaceBorneRadar(radar, carrierTLE, attitudeLaw);
        
        // Load objects catalog
        final URL catalogURL = SimpleChaserTarget.class.getResource("/catalog.tle");
        List<CelestrakTLE> objectsList = CelestrakTLE
        		.parseTLE(catalogURL.getPath().substring(1), OBJECTS_NUMBER);      
        
        // Build propagator        
		final AbsoluteDate start = new AbsoluteDate(YEAR, MONTH, DAY, HOUR, MIN, SEC, 
				TimeScalesFactory.getUTC());		
		final AbsoluteDate end = start.shiftedBy(PROPAGATION_HORIZON);
        
		VisibilityTLEPropagator propagator = new VisibilityTLEPropagator(
				carrier,
				objectsList);
		
		// Progress display step handler          
        propagator.getTLEPropagator().setStepHandler(0.05*PROPAGATION_HORIZON, 
        		(s) -> {
        					LocalDateTime now = LocalDateTime.now(); 
        					Double progress = (s.getDate().durationFrom(start))/PROPAGATION_HORIZON;
        					System.out.println(dtf.format(now) + " " + progress*100 + "%");
        			});
        
        // Propagate	
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation start");
        propagator.getTLEPropagator().propagate(start, end);
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation end \n");
        
        // Chronogram and synthesis generation
        RadarVisibilityEventLogger.logEvents(propagator.getLogger().getLoggedEvents(), 
        		start, end);
                
        EventSynthesis.generate(propagator.getLogger().getLoggedEvents(), 
        		start, end, 
        		OBJECTS_NUMBER, propagator.getUntreatedTargets());		
	}

}