package MultipleTargetsMonitoring;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.LofOffset;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.detection.VisibilityTLEPropagator;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.LOFType;
import org.orekit.logger.RadarVisibilityEventLogger;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.PropagatorsParallelizer;
import org.orekit.propagation.events.EventsLogger;
import org.orekit.propagation.events.EventsLogger.LoggedEvent;
import org.orekit.propagation.sampling.MultiSatStepHandler;
import org.orekit.propagation.sampling.OrekitStepInterpolator;
import org.orekit.spacecraft.Radar;
import org.orekit.spacecraft.SpaceBorneRadar;
import org.orekit.synthesis.EventSynthesis;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.ChronologicalComparator;
import org.orekit.time.TimeScalesFactory;

import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

import SimpleChaserTarget.SimpleChaserTarget;

public class NThreadParallelPropagation {
	
	static class ParallelStepHandler implements MultiSatStepHandler {

		ParallelStepHandler() {
		}
		
		@Override
		public void handleStep(List<OrekitStepInterpolator> interpolators) {
			
			// Display progression for each thread
			final AbsoluteDate start = new AbsoluteDate(YEAR, MONTH, DAY, HOUR, MIN, SEC, 
					TimeScalesFactory.getUTC());
			
			DateTimeFormatter dtf = DateTimeFormatter.ofPattern("uuuu/MM/dd HH:mm:ss.SSS");
			LocalDateTime now = LocalDateTime.now(); 
			
			String message = dtf.format(now);
			for (OrekitStepInterpolator interpolator : interpolators) {
				AbsoluteDate currentDate = interpolator.getCurrentState().getDate();
				Double progress = (currentDate.durationFrom(start))/PROPAGATION_HORIZON;
				message += " " + progress*100 + "%";
			}
			
			//System.out.println(message);
			
		}
		
	}
	
	// Limit on number of objects to be loaded 
	private static int OBJECTS_NUMBER = 20000;
	
	// Number of propagators to be ran in parallel
	private static int PROPAGATORS_NUMBER = 10;
	
	// Radar input file name
	private static String RADAR_INPUT_FILE = "fixed-radar-constant-attitude.yaml";
	
	// Propagation start date
	private static int YEAR = 2023;
	private static int MONTH = 01;
	private static int DAY = 26;
	private static int HOUR = 03;
	private static int MIN = 35;
	private static double SEC = 45.396;
	
	// Propagation horizon
	private static double PROPAGATION_HORIZON = 3600;
	
	public static void main(final String[] args) 
			throws URISyntaxException, 
					StreamReadException, 
					DatabindException, 
					IOException, 
					InterruptedException {
		
		// Load Orekit data
        final File home = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, "
            		+ "unzip it in %s and rename it 'orekit-data' for this "
            		+ "tutorial to work%n",
                    "orekit-data-master.zip", 
                    "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                    home.getAbsolutePath());
            System.exit(1);
        }
        
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));  
		       
        // Time display
        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("uuuu/MM/dd HH:mm:ss.SSS");                    
        
        // Build propagators        
		final AbsoluteDate start = new AbsoluteDate(
				YEAR, MONTH, DAY, 
				HOUR, MIN, SEC, 
				TimeScalesFactory.getUTC());
		final AbsoluteDate end = start.shiftedBy(PROPAGATION_HORIZON);
        
		final int step = OBJECTS_NUMBER/PROPAGATORS_NUMBER;
		final int lastStep = step + OBJECTS_NUMBER%PROPAGATORS_NUMBER;
		
		System.out.println(">> Building propagators \n"
				+ "Number of objects : " + OBJECTS_NUMBER + "\n"
				+ "Number of propagators required : " 
				+ PROPAGATORS_NUMBER + "\n");
				
		List<Propagator> propagators = new ArrayList<>();
		List<EventsLogger> loggers = new ArrayList<>();
		List<CelestrakTLE> untreatedTargets = new ArrayList<>();
		for (int i = 0; i < PROPAGATORS_NUMBER; i++) {			
	        
	        // Radar input file
	        final URL url = SimpleChaserTarget.class.getResource("/" + RADAR_INPUT_FILE);
	        if (url == null) {
	            System.err.println("Resource " + RADAR_INPUT_FILE + " not found.");
	            System.exit(1);
	        }

			final File inputFile  = new File(url.toURI().getPath());  
			
			 // Load objects catalog
	        final URL catalogURL = SimpleChaserTarget.class.getResource("/catalog.tle");
			
			// Read radar input parameters from parsed file
	        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
	        mapper.findAndRegisterModules();
	        final Radar radar = mapper.readValue(inputFile, Radar.class);
	        radar.setOrientation(radar.getU1(), radar.getU2(), radar.getU3());
			
			// Parse radar carrier parameters    
	        final URL carrierURL = SimpleChaserTarget.class.getResource("/carrier.tle");
	        CelestrakTLE carrierTLE = CelestrakTLE.parseTLE(
	        		carrierURL.getPath().replace("/", "\\").substring(1), null).get(0);

			// Load objects catalog
			List<CelestrakTLE> objectsList = CelestrakTLE
	        		.parseTLE(catalogURL.getPath().substring(1), OBJECTS_NUMBER);   
			
			int startIndex = i*step;
			int endIndex = startIndex + step;
			if (i == PROPAGATORS_NUMBER - 1) 
				endIndex = startIndex + lastStep;
			
	        // Align satellite frame with TNW frame
	        final AttitudeProvider attitudeLaw = new LofOffset(FramesFactory.getTEME(), LOFType.TNW);
	        
	        // Radar carrier
	        SpaceBorneRadar carrier = 
	        		new SpaceBorneRadar(radar, carrierTLE, attitudeLaw);
			
	        List<CelestrakTLE> currentList = objectsList.subList(startIndex, endIndex);
	        VisibilityTLEPropagator current = new VisibilityTLEPropagator(
					carrier,
					currentList);
			
			for (CelestrakTLE item : objectsList.subList(startIndex, endIndex)) {
				System.out.println(item.getName());
			}
			
			propagators.add(current.getTLEPropagator());
			loggers.add(current.getLogger());
			untreatedTargets.addAll(current.getUntreatedTargets());
			
			System.out.println("Propagator for objects " 
			+  (startIndex+1) + " to " + (endIndex) 
			+ " (listed above) built successfully \n");
			
		}		
		
		ParallelStepHandler multiStepHandler = new ParallelStepHandler();
		PropagatorsParallelizer pp = 
				new PropagatorsParallelizer(propagators, multiStepHandler);
        
        // Propagate	
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation start");
		pp.propagate(start, end);
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation end \n");
        
        // Chronogram and synthesis generation
        List<LoggedEvent> loggedEvents = new ArrayList<>();
        for (EventsLogger logger : loggers) {
        	loggedEvents.addAll(logger.getLoggedEvents());           
        }
        
        Collections.sort(loggedEvents, new ChronologicalComparator());
        RadarVisibilityEventLogger.logEvents(loggedEvents,
        		start, end);
                
        EventSynthesis.generate(loggedEvents, 
        		start, end, 
        		OBJECTS_NUMBER, untreatedTargets);
		
	}

}