package MultipleTargetsMonitoring;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.LofOffset;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.detection.VisibilityTLEPropagator;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.LOFType;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.PropagatorsParallelizer;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.propagation.sampling.MultiSatStepHandler;
import org.orekit.propagation.sampling.OrekitStepInterpolator;
import org.orekit.spacecraft.Radar;
import org.orekit.spacecraft.SpaceBorneRadar;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScalesFactory;

import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

import SimpleChaserTarget.SimpleChaserTarget;

public class TwoThreadParallelPropagation {
	
	static class ParallelStepHandler implements MultiSatStepHandler {

		ParallelStepHandler() {
		}
		
		@Override
		public void handleStep(List<OrekitStepInterpolator> interpolators) {
			
			final AbsoluteDate start = new AbsoluteDate(YEAR, MONTH, DAY, HOUR, MIN, SEC, 
					TimeScalesFactory.getUTC());
			
			DateTimeFormatter dtf = DateTimeFormatter.ofPattern("uuuu/MM/dd HH:mm:ss.SSS");
			LocalDateTime now = LocalDateTime.now(); 
			
			AbsoluteDate currentDate1 = interpolators.get(0).getCurrentState().getDate();
			AbsoluteDate currentDate2 = interpolators.get(1).getCurrentState().getDate();
			
			Double progress1 = (currentDate1.durationFrom(start))/PROPAGATION_HORIZON;
			Double progress2 = (currentDate2.durationFrom(start))/PROPAGATION_HORIZON;
			
			System.out.println(dtf.format(now) + " " + progress1*100 + "%" 
					+ " " + progress2*100 + "%");
			
		}
		
	}
	
	// Limit on number of objects to be loaded 
	private static int OBJECTS_NUMBER = 300;
	
	// Number of propagators to be ran in parallel
	private static int PROPAGATORS_NUMBER = 2;
	
	// Radar input file name
	private static String RADAR_INPUT_FILE = "fixed-radar-constant-attitude.yaml";
	
	// Propagation start date
	private static int YEAR = 2023;
	private static int MONTH = 01;
	private static int DAY = 26;
	private static int HOUR = 03;
	private static int MIN = 35;
	private static double SEC = 45.396;
	
	// Propagation horizon
	private static double PROPAGATION_HORIZON = 3600;
	
	public static void main(final String[] args) 
			throws URISyntaxException, 
					StreamReadException, 
					DatabindException, 
					IOException, 
					InterruptedException {
		
		// Load Orekit data
        final File home = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, "
            		+ "unzip it in %s and rename it 'orekit-data' for this "
            		+ "tutorial to work%n",
                    "orekit-data-master.zip", 
                    "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                    home.getAbsolutePath());
            System.exit(1);
        }
        
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));  
		       
        // Time display
        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("uuuu/MM/dd HH:mm:ss.SSS");
        
        final URL url = SimpleChaserTarget.class.getResource("/" + RADAR_INPUT_FILE);
        if (url == null) {
            System.err.println("Resource " + RADAR_INPUT_FILE + " not found.");
            System.exit(1);
        }

		final File inputFile  = new File(url.toURI().getPath());                      
        
        // Build propagators        
		final AbsoluteDate start = new AbsoluteDate(YEAR, MONTH, DAY, HOUR, MIN, SEC, 
				TimeScalesFactory.getUTC());
		final AbsoluteDate end = start.shiftedBy(PROPAGATION_HORIZON);        
		
		System.out.println(">> Building propagators \n"
				+ "Number of objects : " + OBJECTS_NUMBER + "\n"
				+ "Number of propagators required : " 
				+ PROPAGATORS_NUMBER + "\n");
				
		List<Propagator> propagators = new ArrayList<>();
			
		// Load objects catalog
        final URL catalogURL = SimpleChaserTarget.class.getResource("/catalog.tle");
		
		// Read radar input parameters from parsed file
        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        final Radar radar1 = mapper.readValue(inputFile, Radar.class);
        radar1.setOrientation(radar1.getU1(), radar1.getU2(), radar1.getU3());
        final Radar radar2 = mapper.readValue(inputFile, Radar.class);
        radar2.setOrientation(radar2.getU1(), radar2.getU2(), radar2.getU3());
		
		// Parse radar carrier parameters    
        final URL carrierURL = SimpleChaserTarget.class.getResource("/carrier.tle");
        CelestrakTLE carrierTLE1 = CelestrakTLE.parseTLE(
        		carrierURL.getPath().replace("/", "\\").substring(1), null).get(0);
        CelestrakTLE carrierTLE2 = CelestrakTLE.parseTLE(
        		carrierURL.getPath().replace("/", "\\").substring(1), null).get(0);

		// Load objects catalog
		List<CelestrakTLE> objectsList1 = CelestrakTLE
        		.parseTLE(catalogURL.getPath().substring(1), OBJECTS_NUMBER);  
		List<CelestrakTLE> objectsList2 = CelestrakTLE
        		.parseTLE(catalogURL.getPath().substring(1), OBJECTS_NUMBER);  
		
        // Align satellite frame with TNW frame
        final AttitudeProvider attitudeLaw1 = new LofOffset(FramesFactory.getTEME(), LOFType.TNW);
        final AttitudeProvider attitudeLaw2 = new LofOffset(FramesFactory.getTEME(), LOFType.TNW);
        
        // Radar carrier
        SpaceBorneRadar carrier1 = 
        		new SpaceBorneRadar(radar1, carrierTLE1, attitudeLaw1);
        SpaceBorneRadar carrier2 = 
        		new SpaceBorneRadar(radar2, carrierTLE2, attitudeLaw2);
		
        List<CelestrakTLE> currentList1 = objectsList1.subList(0, OBJECTS_NUMBER/2);
        List<CelestrakTLE> currentList2 = objectsList2.subList(OBJECTS_NUMBER/2, OBJECTS_NUMBER);
		TLEPropagator current1 = new VisibilityTLEPropagator(
				carrier1,
				currentList1)
				.getTLEPropagator();
		TLEPropagator current2 = new VisibilityTLEPropagator(
				carrier2,
				currentList2)
				.getTLEPropagator();
		
		propagators.add(current1);
		for (CelestrakTLE item : objectsList1.subList(0, OBJECTS_NUMBER/2)) {
			System.out.println(item.getName());
		}
		System.out.println("Propagator for objects 1 to " + OBJECTS_NUMBER/2 
				+ " (listed above) built successfully \n");
		
		propagators.add(current2);	
		for (CelestrakTLE item : objectsList1.subList(0, OBJECTS_NUMBER/2)) {
			System.out.println(item.getName());
		}
		System.out.println("Propagator for objects " + (OBJECTS_NUMBER/2 + 1)
				+ " to " + OBJECTS_NUMBER
				+ " (listed above) built successfully \n");
		
		ParallelStepHandler multiStepHandler = new ParallelStepHandler();
		PropagatorsParallelizer pp = 
				new PropagatorsParallelizer(propagators, multiStepHandler);
        
        // Propagate	
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation start");
		pp.propagate(start, end);
        System.out.println(dtf.format(LocalDateTime.now()) + " Propagation end \n");
		
	}

}