package SimpleChaserTarget;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.hipparchus.util.FastMath;
import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.LofOffset;
import org.orekit.bodies.CelestialBodyFactory;
import org.orekit.celestrak.CelestrakTLE;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.detection.VisibilityTLEPropagator;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.LOFType;
import org.orekit.logger.RadarVisibilityEventLogger;
import org.orekit.spacecraft.Radar;
import org.orekit.spacecraft.SpaceBorneRadar;
import org.orekit.synthesis.EventSynthesis;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScalesFactory;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

public class SimpleChaserTarget {
	
	private static int OBJECTS_NUMBER = 1;
	
	private static String RADAR_INPUT_FILE = "in-plane-radar.yaml";
	
	private static double PROPAGATION_HORIZON = 86400;
	
	public static void main(final String[] args) 
			throws IOException, URISyntaxException, InterruptedException {
		
		// Load Orekit data
        final File home       = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                              orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, "
            		+ "unzip it in %s and rename it 'orekit-data' for this tutorial to work%n",
                    "orekit-data-master.zip", 
                    "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                              home.getAbsolutePath());
            System.exit(1);
        }
        
        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));      
		        
        // Parse radar inputs from file
        System.out.println(">> Reading radar input file");
        
        final URL url = SimpleChaserTarget.class.getResource("/" + RADAR_INPUT_FILE);
        if (url == null) {
            System.err.println("Resource " + RADAR_INPUT_FILE + " not found.");
            System.exit(1);
        }

		final File inputFile  = new File(url.toURI().getPath());
        
        // Read radar input parameters from parsed file
        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        final Radar radar = mapper.readValue(inputFile, Radar.class);
        radar.setOrientation(radar.getU1(), radar.getU2(), radar.getU3());
        
        System.out.println("File " + inputFile.getName() 
        		+ " parsed successfully. Properties : \n" 
        		+ "  Name : " + radar.getName() + "\n"
                + "  Range : " + radar.getRange() + " m \n"
                + "  Half aperture : " + radar.getHalfAperture() + " deg \n"
                + "  Orientation : " + radar.getOrientation() + " \n");   
		
        // Parse radar carrier parameters    
        final URL carrierURL = SimpleChaserTarget.class.getResource("/carrier.tle");
        CelestrakTLE carrierTLE = CelestrakTLE.parseTLE(
        		carrierURL.getPath().replace("/", "\\").substring(1), null).get(0);
        
        Double mu = CelestialBodyFactory.getEarth().getGM();
        Double carrierN = carrierTLE.getMeanMotion();
        Double carrierA = FastMath.cbrt(mu/(carrierN*carrierN));
        System.out.println("Carrier semi-major axis from TLE input : " 
                + carrierA + " m. \n");               
        
        // Align satellite frame with TNW frame
        final AttitudeProvider attitudeLaw = new LofOffset(FramesFactory.getTEME(), LOFType.TNW);
        
        // Radar carrier
        SpaceBorneRadar carrier = 
        		new SpaceBorneRadar(radar, carrierTLE, attitudeLaw);

        // Target
        System.out.println(">> Building target from carrier TLE");
        Double shiftA = -100000.0;
        Double angularShiftRate = (carrierA + shiftA)*
        		computeAngularShiftRate(carrierA, FastMath.abs(shiftA));
        
        System.out.println("Shift from carrier semi-major axis : " + shiftA + " m \n" 
        		+ "Associated angular shift rate : " 
        		+ angularShiftRate + " m/s.");
        
        Double durationInFov = radar.getRange()/angularShiftRate;
        Double estimatedLap = 2*FastMath.PI*carrierA/angularShiftRate;
        
        System.out.println("Estimated first exit from radar field of view : " 
        		+ carrierTLE.getDate().shiftedBy(durationInFov) + "\n"
        		+ "Estimated lap duration : " 
        		+ estimatedLap + " s (" + estimatedLap/86400 + " days). \n");
        
        CelestrakTLE targetTLE = new CelestrakTLE(
        		"Target", 
				carrierTLE.getLine1(),
        		carrierTLE.getLine2().replace(
                		carrierTLE.getLine2().substring(52, 63), 
                		computeMeanMotion(carrierA + shiftA).toString().substring(0, 11)));
        		
        System.out.println("Target TLE shifted from carrier TLE : \n" 
        		+ targetTLE.toString() + "\n");      
        
        List<CelestrakTLE> targetList = new ArrayList<CelestrakTLE>(); 
        targetList.add(targetTLE);
        
        // Propagation parameters
        AbsoluteDate start = new AbsoluteDate(2023, 01, 10, 03, 35, 35.396, TimeScalesFactory.getUTC());
        AbsoluteDate absoluteEnd = new AbsoluteDate(2023, 01, 18, 18, 00, 00, TimeScalesFactory.getUTC());
        double absoluteHorizon = absoluteEnd.durationFrom(start);
        AbsoluteDate end = start.shiftedBy(PROPAGATION_HORIZON);        
        
        // Build propagator
        VisibilityTLEPropagator propagator = new VisibilityTLEPropagator(
				carrier, 
				targetList);
        
        // Propagate
        propagator.getTLEPropagator().propagate(start.shiftedBy(PROPAGATION_HORIZON));
        
        // Chronogram and synthesis generation
        RadarVisibilityEventLogger.logEvents(propagator.getLogger().getLoggedEvents(), 
        		start, end);
                
        EventSynthesis.generate(propagator.getLogger().getLoggedEvents(), 
	    		start, end, OBJECTS_NUMBER, propagator.getUntreatedTargets());
		
	}
	
	public static Double computeMeanMotion(Double a) {
		return FastMath.sqrt(CelestialBodyFactory.getEarth().getGM()/FastMath.pow(a,  3))
				*86400/(2*FastMath.PI);
	}
	
	public static Double computeAngularShift(Double a, double da, double dt) {
		
		return computeAngularShiftRate(a, da)*dt;
		
	}
	
    public static Double computeAngularShiftRate(Double a, double da) {
    	
		return 1.5*(FastMath.sqrt((CelestialBodyFactory.getEarth().getGM()))/
				(FastMath.pow(a, 2)*FastMath.sqrt(a))*da);
		
	}
	
}
